/* dateshift.c 
 * 
 * Date shifting for processes
 */


/* The necessary header files */

/* Standard in kernel modules */
#include <linux/kernel.h>   /* We're doing kernel work */
#include <linux/module.h>   /* Specifically, a module */
#include <linux/sched.h>    /* For task struct and 'current' */
#include <linux/string.h>   /* String library */
#include <linux/fs.h>

#if CONFIG_MODVERSIONS==1
#define MODVERSIONS
#include <linux/modversions.h>
#endif        

#include <sys/syscall.h>  /* The list of system calls */

/* For the current (process) structure, we need
 * this to know who the current user is. */
#include <linux/mm.h>
#include <linux/timex.h>
#include <linux/smp_lock.h>  


#ifndef KERNEL_VERSION
#define KERNEL_VERSION(a,b,c) ((a)*65536+(b)*256+(c))
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
#include <asm/uaccess.h>
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)
#include "ltstring.h"
#endif


MODULE_DESCRIPTION("System time shift module");

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,4,9)
MODULE_LICENSE("GPL");
#endif


extern void *sys_call_table[];
#define KMALLOC_MAXSIZE		131072UL	/* Maximum allowed request size */
extern void *kmalloc(size_t, int);
extern void kfree(const void *);

/* Pointers to the original system calls */
asmlinkage int (*original_call)(time_t *);
asmlinkage void (*original_exit_call)(int);

/* Config params list */
struct param_list {
    char name[20];
    long timeshift;
    int mod;
    int i_start;
    int i_end;
    struct param_list *next;
    };
    
/* Active shifted process list */    
struct ps_list {
    int pid; /* 0 - for free node */
    char name[20];
    int i_count;
    struct ps_list *next;
    };
    
/* Heads and tails for the structures above */
static struct param_list *param_list_head = NULL;
static struct param_list *param_list_tail = NULL;

static struct ps_list *ps_head = NULL;        
static struct ps_list *ps_tail = NULL;        


/* Our time syscall */
asmlinkage int my_tyme(time_t *tm)
{
    struct param_list *p;
    struct ps_list *ps, *new_ps;
    time_t real_time;
  
    /* Call the original sys_time */
    real_time = original_call((time_t *)tm);

    /* Check current process against param_list
     * to see whether we manage it */
   
    for (p = param_list_head; p ; p = p->next) {
        if ( strcmp(current->comm, p->name) == 0 ) {
            /* find it in the ps_list process list */
  	    for (ps = ps_head; ps; ps = ps->next) {
	        if (current->pid == ps->pid
	            && strcmp(current->comm, ps->name) == 0) {
	            if ( ((ps->i_count >= p->i_start
	  	           && ps->i_count < p->i_end)
		              || p->i_start == 0)) goto shift_it;		    
	            else {
		        (ps->i_count)++;
		        return real_time;
		    }
		}
	    }
	    /* if we are here - there is no ps_list entry 
	     * for this process yet. Create it */
	     
            /* First seek for available node in ps_list */
  	    new_ps = NULL;
	    for (ps = ps_head; ps; ps = ps->next) {
	        if (ps->pid == 0) new_ps = ps;
	    }
	    ps = new_ps;
	    if (!ps) {  /* free ps_list is not available - allocate new one */
	        if ((ps = (struct ps_list *) kmalloc(sizeof(struct ps_list), GFP_KERNEL) )) {
		    if (!ps_head) ps_head = ps;
		    if (ps_tail) ps_tail->next = ps;
		    ps_tail = ps;
		}
	        else {
    	  	    printk("do_param: Not enough Kernel memory\n");
		    return real_time;
	        }
	    }
	    strcpy(ps->name, current->comm);
	    ps->pid = current->pid;
	    if (!new_ps) ps->next = NULL;
	    ps->i_count = 0; 	
	    if (p->i_start <= 1) goto shift_it; /* either first or always */
	    else ps->i_count++;
	    break;
	}
    }	  

    return real_time;
  
shift_it:

    ps->i_count++;

    switch (p->mod) {
        case '=': 
            return (time_t) p->timeshift;
        case '+': 
            return real_time + (time_t) p->timeshift;
        case '-':
            return real_time - (time_t) p->timeshift;
    }
    
    return real_time;
}

/* Our exit call */
asmlinkage void my_exit(int ret)
{
    struct ps_list *ps;
 
    /* mark ps_list entry as free to reuse */
    for (ps = ps_head; ps; ps = ps->next) {
        if (current->pid == ps->pid) { ps->pid = 0; break; }
    }

    return original_exit_call(ret);
}

/* Integer character string to Integer conversion */
static int my_atoi(char *name)
{
    int val = 0;

    for (;; name++) {
	switch (*name) {
	    case '0'...'9':
		val = 10 * val + (*name - '0');
		break;
	    default:
		return val;
	}
    }
}

/* Trim all blanks before and after actual information */
char *strtrim(char *str)
{
    char *p, *pp;

    for (p = str; *p != '\0'; ++p) {
        if (*p != ' ') break;
    }
    for (pp = p; *pp != '\0'; ++pp) {
        if (*pp == ' ') { *pp = '\0'; return p; }
    }
    return p;
}    
    

/* Parse one config line */
void do_parse(char *line)
{
    int tok = 1;
    int sh;
    int mod = '=';
    long timeshift = 0L;
    char *name, *p;
    int i_start = 0, i_end = 0;
    char *t_start, *t_end;
    struct param_list *np;
    
    
    /* for the case they separate tokens by tabs */
    for (p = line; *p != '\0'; ++p) if (*p == '\t') *p = ' ';
    /* split into the tokens */
    t_start = line;
    t_end = strtok(line, " ");
    /* Only 4 tokens are for our interst */
    for ( tok = 1; t_end && tok <= 4; t_end = strtok(NULL, " "), t_start = t_end, tok++ ) {
	t_start = strtrim(t_start);
        if (*t_start == '#') break;  /* strip comments */
	switch (tok) {
	    case 1:
		name = t_start;
		break;
	    case 2:
	        sh = 0;
		mod = ' ';
		if (*t_start == '=') { sh++; mod = '='; }
		if (*t_start == '+') { sh++; mod = '+'; }
		if (*t_start == '-') { sh++; mod = '-'; }
                timeshift = (long) my_atoi(t_start+sh);
		if (timeshift == 0) { 
		    printk("dateshift: Error, parsing configuration file - timeshift value is incorrect\n"); 
		    return; 
		}
		break;
	    case 3:
	        i_start = my_atoi(t_start);
		/* fall */
	    case 4:
	        i_end = my_atoi(t_start);
		break;
	    default:
	        break;
	}
    
    }

    if (tok < 3) return;

    /* If we are here - everything's Ok - then add param_list entry */
     
#ifdef DEBUG
	printk("%s\t%c%ld\t%d\t%d\n", name,
	       mod, timeshift,
	       i_start, i_end); 
#endif	

    /* allocate new new_param structure
     * copy all parameters to it
     * and link it to the list */
    if ((np = (struct param_list *) kmalloc(sizeof(struct param_list), GFP_KERNEL) )) {
	if (!param_list_head) param_list_head = np;
	if (param_list_tail)
	    param_list_tail->next = np;
	param_list_tail = np;
	strncpy(np->name, name, 20); /* to prevent buffer override */
	*(np->name+20) = '\0';
	np->mod = mod;
	np->i_start = i_start; 	
	np->i_end = i_end;
	np->timeshift = (long) timeshift; 	
	np->next = NULL;
    }
    else {
	printk("do_param: Not enough Kernel memory\n");
    }

}

/* Initialize the module - replace system calls,
 * read configuration file "/etc/dateshift.conf" into memory 
 * and initialize memory structures */
#define PAGE_COUNT 10

int init_module()
{
    struct file *f = NULL;
    const char *filename = "/etc/dateshift.conf";
    int count = 0, size = 0;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
    mm_segment_t old_fs;
    loff_t pos = 0;
#endif
    char *buffer = NULL, *point = NULL, *offset;
#ifdef DEBUG
    struct param_list *p;
#endif    

/*    f = open_exec(filename);*/
    f = filp_open(filename, O_RDONLY, 0);
    if (IS_ERR(f)) {
        printk("\nConfiguration file /etc/dateshift.conf not found\nNothing to do!\n");
        return -2;
	}
    buffer = (char *) kmalloc(PAGE_COUNT * PAGE_SIZE, GFP_KERNEL);
    do {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
        count = kernel_read(f, 0, buffer+count, PAGE_SIZE);
#else
	old_fs = get_fs();
	set_fs(get_ds());
	count = f->f_op->read(f, buffer+count, PAGE_SIZE, &pos);
	set_fs(old_fs);
#endif
        size += count;
    }
    while (count == PAGE_SIZE && count != 0);

    filp_close(f, NULL);
  
    *(buffer+size) = '\0';
  
#ifdef DEBUG
    printk("%d\n%s\n", size, buffer);
#endif

    /* Parse config file */
    offset = buffer;
    while ( (point = strchr(offset, '\n')) ) {
        *point = '\0';
        do_parse(offset);
        offset = point+1;
    }
    do_parse(offset);  /* parse the last line */
#ifdef DEBUG
    for (p = param_list_head; p ; p = p->next) {
	printk("%s  %c%ld  %d  %d\n", p->name,
	       p->mod, p->timeshift,
	       p->i_start, p->i_end); 
    }        
#endif  
    /* Free unnecessary 'buffer' memory */
    if (buffer) kfree(buffer);

    /* Keep a pointer to the original 'time' function in 
     * original_call, and then replace the system call 
     * in the system call table with my_time.
     * Same for the 'exit' function */
    original_call = sys_call_table[__NR_time];
    sys_call_table[__NR_time] = my_tyme;

    original_exit_call = sys_call_table[__NR_exit];
    sys_call_table[__NR_exit] = my_exit;

    return 0;
}


/* Cleanup - unregister module */
void cleanup_module()
{
    struct param_list *p, *pn;
    struct ps_list *ps, *psn;

    /* Free memory from param_list and ps_list */
    p = param_list_head;
    while (p) {
        pn = p;
        p = p->next;
        if (pn) kfree(pn);
    }        

    ps = ps_head;
    while (ps) {
        psn = ps;
        ps = ps->next;
        if (psn) kfree(psn);
    }        

  /* Return the system call back to normal */

  sys_call_table[__NR_time] = original_call;
  sys_call_table[__NR_exit] = original_exit_call;
}  

